import { Worker } from 'worker_threads'
import genericPool from 'generic-pool'
import os from 'os'

import logger from './logger'
// import { QuestionDb } from '../types/basicTypes'

// ---------------------------------------------------------------------------

const alertOnPendingCount = 10
const workerFile = './src/utils/classes.ts'
let pool: any = null
let workers: any = null

// ---------------------------------------------------------------------------

function handleWorkerError(client, err) {
  logger.Log('resourcePromise error', logger.GetColor('redbg'))
  console.error(err)
  pool.destroy(client)
}

export function doALongTask(obj: any): Promise<any> {
  if (pool.pending > alertOnPendingCount) {
    logger.Log(
      `More than ${alertOnPendingCount} callers waiting for free resource! (${pool.pending})`,
      logger.GetColor('redbg')
    )
  }

  return new Promise((resolve) => {
    let currClient
    pool
      .acquire()
      .then(function(client) {
        currClient = client
        doSomething(client, obj)
          .then((res) => {
            resolve(res)
            // TODO: check if result is really a result, and want to release port
            pool.release(client)
            // console.log('[RELEASE]: #' + client.index)
          })
          .catch(function(err) {
            handleWorkerError(currClient, err)
          })
      })
      .catch(function(err) {
        handleWorkerError(currClient, err)
      })
  })
}

export function initWorkerPool(initData: any): void {
  if (workers && pool) {
    logger.Log('WORKER AND POOL ALREADY EXISTS', logger.GetColor('redbg'))
    return
  }
  workers = []
  const factory = {
    create: function() {
      const currInd = workers.length
      const worker = getAWorker(currInd, initData)
      workers.push(worker)
      return {
        worker: worker,
        index: currInd,
      }
    },
    destroy: function(client) {
      // console.log('[DESTROY]')
      client.worker.terminate()
      // console.log('[DESTROYED] #' + client.index)
    },
  }

  const threadCount = process.env.NS_THREAD_COUNT || os.cpus().length - 1
  if (process.env.NS_THREAD_COUNT) {
    logger.Log(
      `Setting thread count from enviroment variable NS_THREAD_COUNT: '${threadCount}'`,
      logger.GetColor('red')
    )
  }

  const opts = {
    min: threadCount, // minimum size of the pool
    max: threadCount, // maximum size of the pool
    maxWaitingClients: 999,
  }

  pool = genericPool.createPool(factory, opts)
}

// TODO: accuire all workers here, and handle errors so they can be removed if threads exit
export function msgAllWorker(data: any): void {
  workers.forEach((worker) => {
    worker.postMessage(data)
  })
}

// ---------------------------------------------------------------------------

function getAWorker(i, initData) {
  const worker = workerTs(workerFile, {
    workerData: {
      workerIndex: i,
      initData: initData,
    },
  })

  worker.setMaxListeners(50)

  // worker.on('message', (msg) => {
  //   logger.Log(`[MAIN]: Msg from worker #${i}`)
  //   logger.Log(msg)
  // })

  // worker.on('online', () => {
  //   logger.Log(`[THREAD #${i}]: Worker ${i} online`)
  // })

  worker.on('error', (err) => {
    logger.Log('Worker error!', logger.GetColor('redbg'))
    console.error(err)
  })

  worker.on('exit', (code) => {
    // TODO: this is critical, whole server should stop, or child threads should be restarted
    logger.Log(
      `[MAIN]: worker #${i} exit code: ${code}`,
      code === 0 ? logger.GetColor('redbg') : logger.GetColor('green')
    )
  })
  return worker
}

// ---------------------------------------------------------------------------

function doSomething(client, obj) {
  const { /* index, */ worker } = client
  return new Promise((resolve) => {
    // console.log('[ACCUIRE]: #' + index)
    worker.postMessage(obj)
    worker.once('message', (msg) => {
      resolve(msg)
    })
  })
}

const workerTs = (file: string, wkOpts: any) => {
  wkOpts.eval = true
  if (!wkOpts.workerData) {
    wkOpts.workerData = {}
  }
  wkOpts.workerData.__filename = file
  return new Worker(
    `
            const wk = require('worker_threads');
            require('ts-node').register();
            let file = wk.workerData.__filename;
            delete wk.workerData.__filename;
            require(file);
        `,
    wkOpts
  )
}
